﻿UIPanelWindows["RaidTrackerFrame"] = { area = "left", pushable = 1, whileDead = 1 };


-- Globals
KA_RaidTracker_LastWipe = 0;

local KA_RaidTracker_Events = { };
local KA_RaidTracker_AutoBossChangedTime = 0;
local KA_RaidTracker_TimeOffset = 0;
local KA_RaidTracker_GameTimeFrame_Update_Original = nil;

function KARaidTracker:OnLoad( frame )
	RaidTrackerFrameHeaderTitle:SetText( self._addonmeta.title );
	RaidTrackerFrameHeaderVersion:SetText( "v" .. self._addonmeta.version .. "" );
	
	-- Register events
	frame:RegisterEvent("CHAT_MSG_LOOT");
	frame:RegisterEvent("CHAT_MSG_SYSTEM");
	frame:RegisterEvent("RAID_ROSTER_UPDATE");
	frame:RegisterEvent("VARIABLES_LOADED");
	frame:RegisterEvent("UPDATE_MOUSEOVER_UNIT");
	frame:RegisterEvent("ZONE_CHANGED_NEW_AREA");
	frame:RegisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH");
	frame:RegisterEvent("CHAT_MSG_MONSTER_YELL");
	frame:RegisterEvent("CHAT_MSG_MONSTER_EMOTE");
	frame:RegisterEvent("PLAYER_LOGIN");
	frame:RegisterEvent("PLAYER_ENTERING_WORLD");
	frame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	frame:RegisterEvent("UNIT_HEALTH");
	frame:RegisterEvent("UPDATE_INSTANCE_INFO");
	
	SlashCmdList["RAIDTRACKER"] = KA_RaidTracker_CommandOptions;
 	SLASH_RAIDTRACKER1 = "/raidtracker";
 	SLASH_RAIDTRACKER2 = "/rt";
end

function KARaidTracker:OnVariablesLoaded( frame )
	RaidTrackerFrame:CreateTitleRegion()
	RaidTrackerFrame:GetTitleRegion():SetAllPoints()
	KA_RaidTracker_GetGameTimeOffset();
end

function KARaidTracker:OnPlayerLogin( frame )
	local _inst = self._acelib;
	_inst.menuopt = { args = { 
		itemoptions	= {	type = "execute", order = 11, name = _inst.L["Item Options"], desc = _inst.L["Show Item Options"],
					passValue = _inst.instance, func = function(self) self:ExecChatCommandAce("io") end,	},
	} }
end

function KA_RaidTracker_CommandOptions(msg)
	KARaidTracker:CommandOptions(msg)
end

function KARaidTracker:CommandOptions(msg)
	local db = self._store.db;
	local options = self._options;
	local frame = RaidTrackerFrame
	local _, _, command, args = string.find(msg, "(%w+)%s?(.*)");	
	if command then
		command = strlower(command);
	end
		
	if command == "debug" then
		if args=="1" then
			options.DebugMode = 1
		elseif args=="0" then
			options.DebugMode = 0
		end
		self:Print("Debug Mode: ".. ((options.DebugMode and "Enabled") or "Disabled"));
	elseif command == "addwipe" then
		KA_RaidTracker_AddWipe();
	elseif command == "deleteall" then
		db.RaidLog = { };
		options.CurrentRaid = nil;
		self:Print("Deleted "..getn(db.RaidLog).." Raids");
		KA_RaidTracker_Update();
		KA_RaidTracker_UpdateView();
	elseif command == "additem" then
		local linkMatch = "|c%x+|Hitem:[-%d:]+|h%[.-%]|h|r"
		local _, _, item, player = string.find(args, "("..linkMatch..")%s+(%w+)");
		if not item or not player then
			self:Print("AddItem: Must supply an item link and a player name.");
		elseif not frame.selected then
			self:Print("AddItem: There is no raid selected");
		else
			if self:AddLootItemDB(frame.selected,item,player) then
				self:Print("AddItem: Added "..item);
			else
				self:Print("AddItem: Could not add "..item);
			end
		end
		KA_RaidTracker_Update();
		KA_RaidTracker_UpdateView();
	elseif command == "io" then
--		local checks = { "item:(%d+):","(%d+)%s?" };
		local checks = { "item:(%d+):" };
		for i,match in pairs(checks) do
			for itemid in string.gmatch(args,match) do
				itemid = tonumber(itemid);
				if db.ItemOptions[itemid] then
					self:Print(itemid.." is already in the Item Options list.");
				else
					db.ItemOptions[itemid] = {id = itemid};
					self:Print("Added "..itemid.." to the Item Options list.");				
				end
			end
		end
		RT_ItemOptionsFrame:Show();
    elseif command == "options" or command == "o" then
		RaidTracker_OptionsFrame:Show();
	elseif command == "join" or command == "leave" then
		if command == "join" then command = "Join" else command = "Leave" end
		if frame.selected then
			if args and strlen(args) > 0 then
				RT_JoinLeaveFrameNameEB:SetText(args);
			end
			RT_JoinLeaveFrameHeaderTitle:SetText(command);
			RT_JoinLeaveFrame.type = command;
			RT_JoinLeaveFrame.raidid = frame.selected;
			RT_JoinLeaveFrame:Show();
		else
			self:Print(command..": There is no raid selected");
		end
	elseif command then
		self:Print("/rt - Shows the main window.");
		self:Print("/rt options|o - Shows Options window");
		self:Print("/rt io - Shows the ItemOptions window");
		self:Print("/rt io [ITEMLINK|ITEMID]... - Adds items to ItemOptions window");
		self:Print("/rt additem [ITEMLINK] [PLAYER] - Adds a loot item to the selected raid");
		self:Print("/rt join [PLAYER] - Add a player to the selected raid");
		self:Print("/rt leave [PLAYER] - Remmoves a player from the selected raid");
		self:Print("/rt deleteall - Deletes all raids");
		self:Print("/rt debug 1|0 - Enables/Disables debug mode");
		self:Print("/rt addwipe - Adds a Wipe with the current timestamp");
	else
		ShowUIPanel(frame);	
	end
end


function KARaidTracker:OnEventCustom( frame, event, ... )
	local db = self._store.db;
	local options = self._options;
	local util = KarmaUtility
	
	local RaidLog = db.RaidLog;
	local julianne_died = false;

	--self:Debug("Event fired", event);
	if ( RT_UpdateFrame.time and RT_UpdateFrame.time <= 2 ) then
		self:Debug("Queueing", frame and frame:GetName(), event, ...);
		tinsert(KA_RaidTracker_Events, {frame, event, ...} );
		return;
	end
--[[
	if options.AutoBattlegroup == 0 and (GetNumBattlefieldScores() > 0 or IsActiveBattlefieldArena() == 1) then
		if options.CurrentRaid then
			self:DeleteRaidDB(options.CurrentRaid);
			self:Debug("Battlegroup detected, removing raid entry from list.",options.CurrentRaid);
			return;
		else
			self:Debug("Battlegroup detected, skipped event.");
			return;
		end;
	end;
]]

	if event == "CHAT_MSG_MONSTER_YELL" or event == "CHAT_MSG_MONSTER_EMOTE" then -- Sorry, Majordomo needs a little hack
		if arg1 == RaidTracker_L.BossKills_Majordomo_Yell then
			self:Debug("It's domo!");
			event = "CHAT_MSG_COMBAT_HOSTILE_DEATH";
			arg1 = string.gsub(util:ConvertGlobalString(UNITDIESOTHER), "%(%.%+%)", RaidTracker_L.BossKills_Majordomo_BossName);
		elseif arg2 == "Karazhan - Chess, Victory Controller" then
			util:Print("Plz contact Eris at www.mldkp.net that Chess event logging is working (incl language)");
			event = "CHAT_MSG_COMBAT_HOSTILE_DEATH";
			arg1 = string.gsub(util:ConvertGlobalString(UNITDIESOTHER), "%(%.%+%)", RaidTracker_L.BossKills_Chess_Event_BossName);
		elseif arg1 == RaidTracker_L.BossKills_Chess_Event_Yell then
			self:Debug("Chess event");
			event = "CHAT_MSG_COMBAT_HOSTILE_DEATH";
			arg1 = string.gsub(util:ConvertGlobalString(UNITDIESOTHER), "%(%.%+%)", RaidTracker_L.BossKills_Chess_Event_BossName);
		elseif arg1 == RaidTracker_L.BossKills_Julianne_Die_Yell then
			self:Debug("Romulo und Julianne");
			event = "CHAT_MSG_COMBAT_HOSTILE_DEATH";
			arg1 = string.gsub(util:ConvertGlobalString(UNITDIESOTHER), "%(%.%+%)", RaidTracker_L.BossKills_Julianne_BossName);
			julianne_died = true;
		end
	end

	if event == "RAID_ROSTER_UPDATE" or (event == "PARTY_MEMBERS_CHANGED" and GetNumRaidMembers() == 0) then
		local count = self:GetAttendeeCount();
		if count == 0 and options.CurrentRaid then
			self:EndRaidDB();									-- if a raid and no peoples, den mark them all lefted
		elseif count > 1 and not options.CurrentRaid then
			if self:CanAutoCreateLog() then
				self:CreateRaidDB();							-- if not a raid and peoples der and is okies to make raid by self, den makes a raid
			end
		elseif options.CurrentRaid then
			self:UpdateRaidDB();
		end
		
	elseif event == "CHAT_MSG_LOOT" then
		local sPlayer, sLink, iCount, bFilter;				-- job here is to get player, link, item count, and if to engage item filters

		local count = self:GetAttendeeCount();
		if count > 1 and not options.CurrentRaid then
			if self:CanAutoCreateLog() then
				self:CreateRaidDB();							-- if not a raid and peoples der and is okies to make raid by self, den makes a raid
			end
		end
		if not options.CurrentRaid then return end
		
		-- get loot info
		if		string.find(arg1, RaidTracker_L.ReceivesLoot1) then
			_, _, sPlayer, sLink = string.find(arg1, RaidTracker_L.ReceivesLoot1);
		elseif	string.find(arg1, RaidTracker_L.ReceivesLoot2) then
			_, _, sLink = string.find(arg1, RaidTracker_L.ReceivesLoot2);
			sPlayer = YOU;
		elseif	string.find(arg1, RaidTracker_L.ReceivesLoot3) then
			_, _, sPlayer, sLink, iCount = string.find(arg1, RaidTracker_L.ReceivesLoot3);
		elseif	string.find(arg1, RaidTracker_L.ReceivesLoot4) then
			_, _, sLink, iCount = string.find(arg1, RaidTracker_L.ReceivesLoot4);
			sPlayer = YOU;
		end
		if sPlayer and sPlayer == YOU then
			sPlayer = UnitName("player");
		end
		self:Debug("CHAT_MSG_LOOT", "arg1",arg1, "results",sPlayer,sLink,iCount,iStart,iEnd);
		if not sLink or not sPlayer then return; end
		
		self:AddLootItemDB( options.CurrentRaid, sLink, sPlayer, iCount );
			
		KA_RaidTracker_Update();
		KA_RaidTracker_UpdateView();

	elseif event == "UPDATE_MOUSEOVER_UNIT" then	
		if not options.AutoBoss == 1 then return end
		
		local autoboss_unitname = UnitName("mouseover");
		local autoboss_newboss;
		if not UnitIsFriend("mouseover", "player") and not UnitInRaid("mouseover") and not UnitInParty("mouseover") then
			--self:Debug("possible mouseover unit update", autoboss_unitname);
			if RaidTracker_BossUnitTriggers[autoboss_unitname] then
				if RaidTracker_BossUnitTriggers[autoboss_unitname] ~= "IGNORE" then
					autoboss_newboss = RaidTracker_BossUnitTriggers[autoboss_unitname];
					KA_RaidTracker_AutoBossChangedTime = GetTime();
				end
			elseif RaidTracker_BossUnitTriggers["DEFAULTBOSS"] and (options["BossCooldown"] == 0 or (GetTime() > (KA_RaidTracker_AutoBossChangedTime + options["BossCooldown"]))) then
				autoboss_newboss = RaidTracker_BossUnitTriggers["DEFAULTBOSS"];
			else
				autoboss_newboss = nil
			end
			if autoboss_newboss and options.BossName ~= autoboss_newboss then
				options.BossName = autoboss_newboss;
				util:Print("RaidTracker AutoBoss Update: "..autoboss_newboss);
			end
		end
	end
	
	-- end of events that dont require current raid
	if not options.CurrentRaid then
		return 
	
	elseif event == "PLAYER_ENTERING_WORLD" then
		self:UpdateRaidDB();
	
	elseif event == "CHAT_MSG_SYSTEM" then
		local name = UnitName("player");
		if name and name ~= UKNOWNBEING and name ~= UNKNOWN then
			local sDate = self:GetDisplayTime();
			local iStart, iEnd, sPlayer = string.find(arg1, RaidTracker_L.LeftGroup);

			self:Debug("RT CHAT_MSG_SYSTEM: ",arg1,sPlayer);
			
			if sPlayer and sPlayer ~= name and sPlayer ~= UKNOWNBEING and sPlayer ~= UNKNOWN and db.Online[sPlayer] then
				tinsert(RaidLog[options.CurrentRaid].Leave,	{ player = sPlayer, time = sDate, }	);
				db.Online[sPlayer] = nil;
				self:Debug(sPlayer, "LEFT", sDate);
			end
			
			KA_RaidTracker_UpdateView();
			KA_RaidTracker_Update();
		end
		
	elseif ( event == "CHAT_MSG_COMBAT_HOSTILE_DEATH" ) then
		self:Debug("CHAT_MSG_COMBAT_HOSTILE_DEATH",arg1);
		local bosskilled, autoboss_newboss;
		local sDate = self:GetDisplayTime();
		
		for unit in string.gmatch(arg1, util:ConvertGlobalString(UNITDIESOTHER)) do
			self:Debug("CHAT_MSG_COMBAT_HOSTILE_DEATH","unit", unit);
			if options.AutoBoss == 2 then
				if(not db.Online[unit]) then
					if(RaidTracker_BossUnitTriggers[unit]) then
						if(RaidTracker_BossUnitTriggers[unit] ~= "IGNORE") then
							autoboss_newboss = RaidTracker_BossUnitTriggers[unit];
						end
					elseif(RaidTracker_BossUnitTriggers["DEFAULTBOSS"]) then
						autoboss_newboss = RaidTracker_BossUnitTriggers["DEFAULTBOSS"];
					else
						autoboss_newboss = nil
					end
					if(autoboss_newboss and options.BossName ~= autoboss_newboss) then
						options.BossName = autoboss_newboss;
						if (unit == "Hase") then
							unit = "V\195\182lli hat den alten Hoppel Hasen erledigt!"; --Just a Joke
						end
						util:Print("KA_RaidTracker AutoBoss Update: "..autoboss_newboss.." ("..unit..")", 1, 1, 0);
					end
				end
			end
			
			if RaidTracker_BossUnitTriggers[unit] and RaidTracker_BossUnitTriggers[unit] ~= "IGNORE" then	
				local newboss = 1;
				-- Romulo and Julianne Hack
				if RaidTracker_BossUnitTriggers[unit] == "Romulo and Julianne" then
					if (julianne_died == false) then
						return;
					end;
				end
				-- Romulo and Julianne Hack
				bosskilled = RaidTracker_BossUnitTriggers[unit];
				if not RaidLog[options.CurrentRaid].BossKills then
					RaidLog[options.CurrentRaid].BossKills = { };
				end
				-- is the boss already killed?
				for key, val in pairs(RaidLog[options.CurrentRaid].BossKills) do
				    if val.boss == bosskilled then
				        newboss = 0;
				    end
				end
				if newboss == 1 then
					tinsert( RaidLog[options.CurrentRaid].BossKills,
						{
							boss = bosskilled,
							time = sDate,
							attendees = self:GetAttendees((options.LogAttendees==1) or (options.LogAttendees==3), true, options.LogAttendees == 3);
						}
					);
					util:Print("RaidTracker Boss Kills: Set kill time for \""..bosskilled.."\" to "..sDate, 1, 1, 0);
					if options.BossAsk == 1 then
						RT_NextBossFrame:Show();
					end;
					if options.LogGuild == 1 then
						KA_RaidTrackerAddGuild();			      
					end;	      
				end
				--if(not RaidLog[options.CurrentRaid].BossKills[bosskilled]) then
				--	RaidLog[options.CurrentRaid].BossKills[bosskilled] = sDate;
				--	util:Print("KA_RaidTracker Boss Kills: Set kill time for \""..bosskilled.."\" to "..sDate, 1, 1, 0);
				--end
			end
		end
		
	elseif event == "UNIT_HEALTH" then
		if options.WipeAsk == 0 or InCombatLockdown() or GetTime()<KA_RaidTracker_LastWipe then return end;
		
		local memberdead, count = 0, self:GetAttendeeCount();
		for i = 1, count, 1 do
			if UnitIsDeadOrGhost(self:GetAttendeeUnitId(i)) then
				memberdead = memberdead + 1;
			end
		end
		if memberdead == count then
			KA_RaidTracker_AddWipe();
			RT_AcceptWipeFrame:Hide();
		elseif (memberdead / count) > options.WipePercent then
			RT_AcceptWipeFrame:Show();
		end;

	elseif event == "ZONE_CHANGED_NEW_AREA" then
		if options.AutoZone == 1 then
			self:DoZoneCheck();
			KA_RaidTracker_DoRaidIdCheck();
		end
		
	elseif ( event == "UPDATE_INSTANCE_INFO" ) then
		KA_RaidTracker_DoRaidIdCheck();
	end
	
end


-- unifies attendees info lookup and raid/party/bgs/arenas

function KARaidTracker:GetAttendeesType( )
	if IsActiveBattlefieldArena() == 1 then return "Arena" end
	if GetNumBattlefieldScores() > 0 then return "Battlegroup" end
	if GetNumRaidMembers() > 0 then return "Raid" end
	if GetNumPartyMembers() > 0 then return "Party" end
	return "Solo"
end


function KARaidTracker:CanAutoCreateLog( )
	local o, t = self._options, self:GetAttendeesType();
	return (t=="Arena" and o.AutoAreana==2) or (t=="Battlegroup" and o.AutoBattlegroup==2) or 
		   (t=="Raid" and o.AutoRaid==2) or (t=="Party" and o.AutoParty==2) or (o.AutoSolo==2);
end

function KARaidTracker:GetAttendeeCount( )		-- if party logging is off then will return 0 if on a party
	local o, t = self._options, self:GetAttendeesType();
	local options = self._options;
	if GetNumRaidMembers() > 0 and (options.AutoRaid >= 1 or
		(GetNumBattlefieldScores() > 0 and options.AutoBattlegroup >= 1)) then
		return GetNumRaidMembers()
	end
	if GetNumPartyMembers() > 0 and (options.AutoParty >= 1 or
		(IsActiveBattlefieldArena() == 1 and options.AutoAreana >= 1)) then
		return GetNumPartyMembers() + 1
	end
	if options.AutoSolo >= 1 then return 1 end	
	return 0;
end

function KARaidTracker:GetAttendeeUnitId( i )
	if tonumber(i)==nil then return i end
	if GetNumRaidMembers() > 0 then return "raid"..i end
	return (i > GetNumPartyMembers()) and "player" or ("party"..i);
end

function KARaidTracker:GetAttendeeInfo( i, isExtended )
	local unit = self:GetAttendeeUnitId(i);
	
	local name, online, zone, level, class;
	if GetNumRaidMembers() > 0 and tonumber(i) then		
		name,_,_,level,_,class,zone,online = GetRaidRosterInfo(i);	-- name, rank, subgroup, level, class, fileName, zone, online = GetRaidRosterInfo(i);
	else
		name,online,zone = UnitName(unit),UnitIsConnected(unit),self:UnitZone(unit);
	end
	if not isExtended then
		return name, online, zone;
	end	
	
	level = level or UnitLevel(unit);
	if not class then _, class = UnitClass(unit); end
	local sex, guild, _, race = UnitSex(unit), GetGuildInfo(unit), UnitRace(unit);
	return name, online, zone, level, class, sex, guild, race;
end

function KARaidTracker:GetAttendees( isInit, chkOnline, chkZone )
	local options = self._options;
	local count, t = self:GetAttendeeCount(), { };
	if isInit == false or count == 0 then return t end

	for i = 1, count do
		name, online, zone = self:GetAttendeeInfo(i);
		if (name and name ~= UKNOWNBEING and name ~= UNKNOWN) and 		-- check name and itentity
			(not chkZone or not zone or zone==GetRealZoneText()) and	-- should check, haz a zone, zone matches
			(not chkOnline or online) then  							-- should check, is online		
				tinsert(t, i);
		end
	end

	return t;
end


function KARaidTracker:SortRaidTable()
	local db = self._store.db;	
	local func = function( a1, a2 )
			return a1 and a2 and self:GetTimestamp(a1.key) > self:GetTimestamp(a2.key);
		end
	table.sort( db.RaidLog, func );
end

		
function KARaidTracker:UpdateRaidDB( isNew )
	local db = self._store.db;	
	local options = self._options;
	if not options.CurrentRaid then return end
	local t = db.RaidLog[options.CurrentRaid];
	
	-- update player info, join and leave times
	local tAttend, sDate = self:GetAttendees(), self:GetDisplayTime();
	for i, v in pairs(tAttend) do
		local name, online, zone, level, class, sex, guild, race = self:GetAttendeeInfo(v, true);

		if not t.PlayerInfos[name] then
			t.PlayerInfos[name] = { };
		end
		local tInfo = t.PlayerInfos[name];		
		if options.SavePlayerDetail == 1 then
			if race then tInfo.race = race; end
			if class then tInfo.class = class; end
			if sex then tInfo.sex = sex; end
			if level > 0 then tInfo.level = level; end
			if guild then tInfo.guild = guild; end
		end

		if isNew or online ~= db.Online[name] then						-- this will change status for existing player
			if isNew or online then
				tinsert( t.Join,  { player = name, time = sDate } );
			end
			if not online then
				tinsert( t.Leave, {	player = name, time = sDate } );
			end
		end
		
		db.Online[name] = online;
	end
	
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KARaidTracker:CreateRaidDB()
	local db = self._store.db;
	local options = self._options;
	KA_RaidTracker_GetGameTimeOffset();
	local sDate = self:GetDisplayTime();
	
	-- start new raid
	self:EndRaidDB()
	tinsert( db.RaidLog, 1, { 
		Loot = { },
		Join = { },
		Leave = { },
		PlayerInfos = { },
		BossKills = { },
		key = sDate
	} );
	options.CurrentRaid = 1;
	self:SortRaidTable();						-- sort by creation timestamp
	self:UpdateRaidDB( true );					-- setup player info, join and leave times

	if options.AutoZone == 1 then
		self:DoZoneCheck()
	end

	KarmaUtility:Debug("Joined new raid at " .. sDate);
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KARaidTracker:DeleteRaidDB( raidid )
	local db = self._store.db;
	local options = self._options;
	local frame = RaidTrackerFrame;
	
	if frame.selected == raidid then			-- unhook frame if selected raid
		if frame.selected > 1 then
			 frame.selected = frame.selected - 1;
		end
		frame.type = "raids";
	end
	if options.CurrentRaid == raidid then		-- unhook db if current raid
		options.CurrentRaid = nil;
		db.Online = { };
	end
	
	table.remove(db.RaidLog, raidid);

	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KARaidTracker:EndRaidDB()
	local db = self._store.db;
	local options = self._options;
    
    if options.CurrentRaid then
		local sTime = self:GetDisplayTime();
        KarmaUtility:Print("Ending current raid at "..sTime);
        
    	for k, v in pairs(db.Online) do
    		tinsert( db.RaidLog[options.CurrentRaid].Leave, { player = k, time = sTime, } );
    	end
    	if not db.RaidLog[options.CurrentRaid].End then
    		db.RaidLog[options.CurrentRaid].End = sTime;
    	end
    	db.Options.CurrentRaid = nil;

		KA_RaidTracker_UpdateView();
		KA_RaidTracker_Update();
    end

   	db.Online = { };
end


function KARaidTracker:CreateLootItemDB( frame, sLink, sPlayer, iCount, ioGetCost )
	local options = self._options;
	local sColor, sItem, sName = self:GetLinkMeta(sLink);
	local sNote, sCosts, sBoss;

	if not sItem then return end

	if ioGetCost then
		local i = 0;
		local splitted = { [0] = 0, [1] = 0, [2] = 0, [3] = 0 };
		for item in string.gmatch(string.gsub(sItem, "^%s*(.-)%s*$", "%1") .. ":", "([^:]*):?") do
			local trimed = string.gsub(item, "^%s*(.-)%s*$", "%1");
			if string.len(trimed) >= 1 then
				splitted[i] = trimed;
				i = i + 1;
			end
		end
		self:Debug("Splitted", splitted[0], splitted[1], splitted[2], splitted[3]);
		if DKPValues and DKPValues[tostring(splitted[0])] then -- AdvancedItemTooltip
			sCosts = tonumber(DKPValues[tostring(splitted[0])]);
		elseif HDKP_GetDKP then -- HoB_DKP
			sCosts = tonumber(HDKP_GetDKP(splitted[0], splitted[1], splitted[2], splitted[3]));
		end
		if sCosts == 0 then
			sCosts = nil;
		end
	end
		
	if options.AutoBoss >= 1 then
		sBoss = options.BossName;
	end

	local name, link, quality, iLevel, minLevel, class, subclass, maxStack, invtype, icon = GetItemInfo("item:"..sItem);
	if icon then 
		_, _, icon = string.find(icon, "^.*\\(.*)$");
	end
	local lootItem = {
		player = sPlayer,
		item = {						
			c = sColor,
			id = sItem,
			tooltip = (options.SaveTooltips == 1) and self:GetItemTooltip(sItem) or nil,
			name = sName,
			icon = icon,
			count = iCount,
			class = class,
			subclass = subclass,
		},
		costs = sCosts,
		boss = sBoss,
		note = sNote,
		zone = GetRealZoneText(),
		time = self:GetDisplayTime(),
		attendees = self:GetAttendees(options.LogAttendees==2, true),
	}

	return lootItem;
end

function KARaidTracker:AddLootItemDB( raidid, sLink, sPlayer, iCount )
	self:Debug("AddLootItemDB","rid",raidid,"link",sLink,"player",sPlayer)
	local frame = RaidTrackerFrame;
	local db = self._store.db;
	local options = self._options;
	if not raidid then return end
	local items = db.RaidLog[raidid].Loot;
	
	iCount = tonumber(iCount) or 1;
	
	-- check via Options
	local sColor, sItem, sName = self:GetLinkMeta(sLink);
	local rarity = self.ColorToRarityTable[sColor];
	local _, _, itemid = string.find(sItem, "^(%d+):%d+:%d+:%d+");
	local io = db.ItemOptions[tonumber(itemid)];
	local ioGroup, ioCostAsk, ioGetCost;
	if (io and io.Track==0)	or ((not io or not io.Track) and not
		(rarity >= options.LogItems) ) then return; end
	ioGroup = (io and io.Group==1) or ((not io or not io.Group) and
		(options.GroupItems~=0 and rarity <= options.GroupItems) );
	ioGetCost = (io and io.CostGet==1) or ((not io or not io.CostGet) and
		(options.CostGet~=0 and rarity >= options.CostGet) );
	ioCostAsk = (io and io.CostAsk==1) or ((not io or not io.CostAsk) and
		(options.CostAsk~=0 and rarity >= options.CostAsk) );
		
	-- add loot item
	local t;
	if ioGroup then
		for k, v in pairs(items) do
			if ( v.item.name == sName and v.player == sPlayer ) then
				v.item.count = (v.item.count or 0) + iCount;
				t = v;
				break;
			end
		end
	end
	if not t then
		t = self:CreateLootItemDB(frame, sLink, sPlayer, iCount, ioGetCost);
		t.id = #items + 1
		tinsert(items, t.id, t);
		if ioCostAsk then									-- code and idea from tlund
			RT_Buttons:ShowFrame("RT_EditCostFrame", "itemcost", raidid, 1);
		end
	end
	return t;	
end

function KARaidTracker:GetLinkMeta( sLink )
	if not sLink then return end
	local _, _, sColor, sItemStr, sName = string.find(sLink, "|c(%x+)|Hitem:([-%d:]+)|h%[(.-)%]|h|r");
	return sColor, sItemStr, sName;
end

function KARaidTracker:GetDisplayTime( )
	return self:GetTimeDisplay(nil, self._options.Timezone, KA_RaidTracker_TimeOffset);
end

function KARaidTracker:DoZoneCheck( )
	local db = self._store.db;
	local options = self._options;
	if not options.CurrentRaid then return end
	
	local newzone = GetRealZoneText();	
	if not db.RaidLog[options.CurrentRaid].zone then
		local list = {RaidTracker_ZoneTriggers, db.ZoneTriggers};
		for i, t in pairs(list) do
			for k, v in pairs(t) do
				if newzone == k then
					db.RaidLog[options.CurrentRaid].zone = v;
					KA_RaidTracker_Update();
					KA_RaidTracker_UpdateView();
					break;
				end
			end
		end
	end
	return true;
end


--[[ *****************************************************************************************************
	******************************************************************************************************
	******************************************************************************************************
	****************************************************************************************************** ]]
	

local KA_RaidTracker_RarityTable = {
	["ff9d9d9d"] = 1,
	["ffffffff"] = 2,
	["ff1eff00"] = 3,
	["ff0070dd"] = 4,
	["ffa335ee"] = 5,
	["ffff8000"] = 6,
	["ffe6cc80"] = 7,
};



function KA_RaidTracker_GetRaidTitle(raidid, hideid, showzone, shortdate)
	local raidLog = KARaidTrackerDB.RaidLog
	local RaidTitle = "";
	if ( raidLog[raidid] and raidLog[raidid].key ) then
		local _, _, mon, day, year, hr, min, sec = string.find(raidLog[raidid].key, "(%d+)/(%d+)/(%d+) (%d+):(%d+):(%d+)");
		if ( mon ) then
			local months = {
				"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
			};
			if ( not hideid ) then
				RaidTitle = RaidTitle .. "[" .. (getn(raidLog)-raidid+1) .. "] ";
			end
			if ( not shortdate ) then
				RaidTitle = RaidTitle .. months[tonumber(mon)] .. " " .. day .. " '" .. year .. ", " .. hr .. ":" .. min;
			else
				RaidTitle = RaidTitle .. mon .. "/" .. day .. " " .. hr .. ":" .. min;
			end
			if ( showzone and raidLog[raidid].zone) then
				RaidTitle = RaidTitle .. " " .. raidLog[raidid].zone;
			end
			return RaidTitle;
		else
			return "";
		end
	end
	return "";
end


function KA_RaidTracker_GetLootId(raidid, sPlayer, sItem, sTime)
	for key, val in pairs( KARaidTrackerDB.RaidLog[raidid].Loot ) do
		if val.player == sPlayer and val.item.id == sItem and val.time == sTime then
			return key;
		end
	end
end

function KA_RaidTracker_Update(frame)
	local frame = RaidTrackerFrame; 
	local db = KARaidTrackerDB;
	
	-- main frame top buttons
    if db.Options.CurrentRaid then
		RaidTrackerFrameEndRaidButton:Enable();
		RaidTrackerFrameSnapshotButton:Enable();
    else
        RaidTrackerFrameEndRaidButton:Disable();
        RaidTrackerFrameSnapshotButton:Disable();
    end
    if KARaidTracker:GetAttendeeCount() > 0 then
		RaidTrackerFrameNewRaidButton:Enable();
	else
		RaidTrackerFrameNewRaidButton:Disable();
	end;
	if #RT_Buttons.LastPage > 0 then
		RaidTrackerFrameBackButton:Enable();
	else
		RaidTrackerFrameBackButton:Disable();
	end
	
	-- process
	local numEntries = getn(db.RaidLog);
	if numEntries <= 0 then
		frame.type = "raids";
		frame.selected = nil;
		RaidTrackerFrameEmptyFrame:Show();
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFrameItems:Hide();
		RT_DetailScrollFramePlayer:Hide();
		RT_DetailScrollFrameEvents:Hide();
		RaidTrackerFrameDeleteButton:Disable();
		RaidTrackerFrameViewButton:Disable();
	else
		RaidTrackerFrameEmptyFrame:Hide();
		RaidTrackerFrameDeleteButton:Enable();
		frame.type = frame.type or "raids";

		local type = frame.type;
		if type == "raids" or type == "items" then
			frame.selected = frame.selected or 1;
			if not db.RaidLog[frame.selected] then 
				frame.selected = numEntries;
			end
			if getn(db.RaidLog[frame.selected].Loot) == 0 then		
				RaidTrackerFrameViewButton:Disable();
			else
				RaidTrackerFrameViewButton:Enable();
			end
			RaidTrackerFrameViewButton:SetText( (type == "raids" and "View Items") or "View Players");
		elseif type == "player" or type == "playeritems" then
			local hasItem;
			for k, v in pairs(db.RaidLog) do
				local loot = v.Loot;
				for key, val in pairs(loot) do
					if val.player == frame.player then hasItem = 1;	break; end
				end
				if hasItem then	break; end
			end
			if not hasItem then 
				if type == "playeritems" then 
					frame.type = "player";
					KA_RaidTracker_Update();
					KA_RaidTracker_UpdateView();
					return;
				end
				RaidTrackerFrameViewButton:Disable();				
			else
				RaidTrackerFrameViewButton:Enable();
			end
			RaidTrackerFrameDeleteButton:Disable();
			RaidTrackerFrameViewButton:SetText((type == "player" and "View Loot") or "View Raids" );
		elseif type == "itemhistory" or type == "events" then
			RaidTrackerFrameViewButton:Disable();
			RaidTrackerFrameDeleteButton:Disable();
		end
	end

	-- ScrollFrame update
	local numEntries = getn(db.RaidLog);
	FauxScrollFrame_Update(RT_ListScrollFrame, numEntries, 6, 16, nil, nil, nil, RT_ListHighlightFrame, 293, 316 );	
	RT_ListHighlightFrame:Hide();
	for i=1, 6, 1 do
		local line = getglobal("RT_RaidLineFrame" .. i);
		local raidid = i + FauxScrollFrame_GetOffset(RT_ListScrollFrame);
		
		if raidid > numEntries then
			line.raidid = nil;
			line:Hide();
		else
			local raidTitle = KA_RaidTracker_GetRaidTitle(raidid, nil, 1, 1);
			local raidTag = db.RaidLog[raidid].note;
			line.raidid = raidid;
			line:SetText( (raidTitle or "") .. ((raidTag and " (" .. raidTag .. ")") or "") );			
			line:Show();
			
			-- Place the highlight and lock the highlight state
			if ( frame.selected and frame.selected == raidid ) then
				KA_RaidTrackerSkillHighlight:SetVertexColor(1, 1, 0);
				RT_ListHighlightFrame:SetPoint("TOPLEFT", "RT_RaidLineFrame"..i, "TOPLEFT", 0, 0);
				RT_ListHighlightFrame:Show();
				line:LockHighlight();
			else
				line:UnlockHighlight();
			end
		end
	end
	
	local text = KARaidTracker:GetAttendeesType() .. " - " .. KARaidTracker:GetAttendeeCount();
	RaidTrackerFrameHeaderText:SetText( text );
end

function KA_RaidTracker_SelectRaid(id)
	local raidid = id + FauxScrollFrame_GetOffset(RT_ListScrollFrame);
	KA_RaidTracker_GetPage();
	RaidTrackerFrame.selected = raidid;
	--if ( getn(KARaidTrackerDB.RaidLog[raidid].Loot) == 0 or ( RaidTrackerFrame.type and RaidTrackerFrame.type ~= "items" ) ) then
		RaidTrackerFrame.type = "raids";
	--end

	KA_RaidTracker_UpdateView();
	KA_RaidTracker_Update();
end

function KA_RaidTracker_ShowInfo( player )
	KA_RaidTracker_GetPage();

	RaidTrackerFrame.type = "player";
	RaidTrackerFrame.player = player;
	RaidTrackerFrame.selected = nil;
	
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KA_RaidTracker_Delete( frame ) -- raidid, type, typeid )
	local db = KARaidTrackerDB
	local type = frame.type;
	local raid = db.RaidLog[frame.raidid];
	
	if type == "raid" then
		KARaidTracker:DeleteRaidDB(frame.raidid);
		return;
	elseif type == "item" then
		table.remove(raid.Loot, frame.typeid);
		for i,v in pairs(raid.Loot) do
			v.id = i;
		end
	elseif type == "player" then
		for key, val in pairs(raid.Join) do
			if val.player == frame.typeid then
				KarmaUtility:Debug("DELETE", "JOIN", "FOUND PLAYER", key, val.player);
				raid.Join[key] = nil;
			end
		end
		for key, val in pairs(raid.Leave) do
			if val.player == frame.typeid then
				KarmaUtility:Debug("DELETE", "LEAVE", "FOUND PLAYER", key, val.player);
				raid.Leave[key] = nil;
			end
		end
		if id == db.Options.CurrentRaid then
			db.Online[frame.typeid] = nil;
		end
		if raid.PlayerInfos then
			raid.PlayerInfos[frame.typeid] = nil;
		end
	end
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KA_RaidTracker_Sort(tbl, method, way)
	if ( way == "asc" ) then
		table.sort(	tbl,
			function(a1, a2)
				return a1[method] < a2[method];
			end
		);
	else
		table.sort(	tbl,
			function(a1, a2)
				return a1[method] > a2[method];
			end
		);
	end
	return tbl;
end


function KA_RaidTracker_SortPlayerRaids(id)
	local sortOptions = KARaidTrackerDB.SortOptions
	if ( RaidTrackerFrame.type == "itemhistory" ) then
		local table = { "name", "looter" };

		if ( sortOptions["itemhistorymethod"] == table[id] ) then
			if ( sortOptions["itemhistoryway"] == "asc" ) then
				sortOptions["itemhistoryway"] = "desc";
			else
				sortOptions["itemhistoryway"] = "asc";
			end
		else
			sortOptions["itemhistoryway"] = "asc";
			sortOptions["itemhistorymethod"] = table[id];
		end
	else		
		if ( sortOptions["playerraidway"] == "asc" ) then
			sortOptions["playerraidway"] = "desc";
		else
			sortOptions["playerraidway"] = "asc";
		end
	end
	KA_RaidTracker_UpdateView();
end


function KA_RaidTracker_CompareItems(a1, a2)
	-- This function could probably be better, but I can't think of any better way while still maintaining functionality
	local util = KarmaUtility
	local sortOptions = KARaidTrackerDB.SortOptions
	
	local filter, method, way = sortOptions["itemsfilter"], sortOptions["itemmethod"], sortOptions["itemway"];
	if ( RaidTrackerFrame.type == "playeritems" ) then
		filter, method, way = sortOptions["playeritemsfilter"], sortOptions["playeritemmethod"], sortOptions["playeritemway"];
	end

	-- Check to see if it matches the rarity requirements
	KarmaUtility:Debug(a2.item.c);
	if ( KA_RaidTracker_RarityTable[a1.item.c] < filter ) then
		return false;
	elseif ( KA_RaidTracker_RarityTable[a2.item.c] < filter ) then
		return true;
	end

	if ( method == "name" ) then
		local c1, c2 = a1.item.name, a2.item.name;
		if ( c1 == c2 ) then
			c1, c2 = a1.player, a2.player;
		end
		if ( way == "asc" ) then
			return c1 < c2;
		else
			return c1 > c2;
		end
	elseif ( method == "looter" ) then
		local c1, c2 = a1.player, a2.player;
		if ( c1 == c2 ) then
			c1, c2 = KA_RaidTracker_RarityTable[a2.item.c], KA_RaidTracker_RarityTable[a1.item.c];
			if ( c1 == c2 ) then
				c1, c2 = a1.item.name, a2.item.name;
			end
		end
		if ( way == "asc" ) then
			return c1 < c2;
		else
			return c1 > c2;
		end
	elseif ( method == "looted" ) then
		if ( way == "asc" ) then
			return util:GetTimestamp(a1.time) < util:GetTimestamp(a2.time);
		else
			return util:GetTimestamp(a1.time) > util:GetTimestamp(a2.time);
		end
	else
		local c1, c2 = KA_RaidTracker_RarityTable[a1.item.c], KA_RaidTracker_RarityTable[a2.item.c];
		if ( c1 == c2 ) then
			c1, c2 = a1.item.name, a2.item.name;
			if ( c1 == c2 ) then
				c1, c2 = a1.player, a2.player;
			else
				return c1 < c2;
			end
		end
		if ( way == "asc" ) then
			return c1 < c2;
		else
			return c1 > c2;
		end
	end
end


function KA_RaidTracker_SortItemBy(id)
	local sortOptions = KARaidTrackerDB.SortOptions
	local table = { "name", "looted", "looter", "rarity" };
	local prefix = "";
	if ( RaidTrackerFrame.type == "playeritems" ) then
		prefix = "player";
	end
	if ( sortOptions[prefix.."itemmethod"] == table[id] ) then
		if ( sortOptions[prefix.."itemway"] == "asc" ) then
			sortOptions[prefix.."itemway"] = "desc";
		else
			sortOptions[prefix.."itemway"] = "asc";
		end
	else
		sortOptions[prefix.."itemmethod"] = table[id];
		sortOptions[prefix.."itemway"] = "asc";
	end
	KA_RaidTracker_UpdateView();
end

function KA_RaidTracker_SortBy(id)
	local sortOptions = KARaidTrackerDB.SortOptions
	local table = { "name", "join", "leave" };
	if ( sortOptions["method"] == table[id] ) then
		if ( sortOptions["way"] == "asc" ) then
			sortOptions["way"] = "desc";
		else
			sortOptions["way"] = "asc";
		end
	else
		sortOptions["method"] = table[id];
		if ( table[id] ~= "leave" ) then
			sortOptions["way"] = "asc";
		else
			sortOptions["way"] = "desc";
		end
	end
	KA_RaidTracker_UpdateView();
end

function KA_RaidTracker_UpdateView( )
	local frame = RaidTrackerFrame; 
	local db = KARaidTrackerDB;	

	if RaidTrackerFrameEmptyFrame:IsVisible() then return end
	local raidid = frame.selected;
	local raid = db.RaidLog[raidid]	
	if raid == nil then
		raidid = nil;
	end
	
	if frame.type == "events" then
		RaidTrackerFrameView2Button:SetText("View Raid");
	else
		RaidTrackerFrameView2Button:SetText("View Events");
		if not raidid or ( (not raid.BossKills or getn(raid.BossKills) == 0) and (not raid.Events or getn(raid.Events) == 0)) then
			RaidTrackerFrameView2Button:Disable();
		else
			RaidTrackerFrameView2Button:Enable();
		end
	end
	
	if ( frame.type == "raids" or not frame.type ) then
		RT_DetailScrollFramePlayers:Show();
		RT_DetailScrollFramePlayer:Hide();
		RT_DetailScrollFrameItems:Hide();
		RT_DetailScrollFrameEvents:Hide();
		local players = { };
		if ( raid ) then

			local playerIndexes = { };
			for key, val in pairs(raid.Join) do
				if ( val.player ) then
					local id = playerIndexes[val.player];
					local time = KarmaUtility:GetTimestamp(val.time);
					if ( not id or time < players[id]["join"] ) then
						
						if ( playerIndexes[val.player] ) then
							players[id] = {
								["join"] = time,
								["name"] = val.player
							};
						else
							tinsert(players, {
								["join"] = time,
								["name"] = val.player
							});
							playerIndexes[val.player] = getn(players);
						end
					end
					id = playerIndexes[val.player];
					if ( not players[id]["lastjoin"] or players[id]["lastjoin"] < time ) then
						players[id]["lastjoin"] = time;
					end
				end
			end
			for key, val in pairs(raid.Leave) do
				local id = playerIndexes[val.player];
				local time = KarmaUtility:GetTimestamp(val.time);
				if ( id ) then
					if ( ( not players[id]["leave"] or time > players[id]["leave"] ) and time >= players[id]["lastjoin"] ) then
						players[id]["leave"] = time;
					end
				end
			end
			for k, v in pairs(players) do
				if ( not v["leave"] ) then
					-- Very ugly hack, I know :(
					players[k]["leave"] = 99999999999;
				end
			end
			players = KA_RaidTracker_Sort(players, db.SortOptions["method"], db.SortOptions["way"]);
			getglobal("RT_DetailScrollFramePlayers").raidid = raidid;
			getglobal("RT_DetailScrollFramePlayers").players = players;
			KA_RaidTracker_DetailScrollFramePlayers_Update();
		end
		KA_RaidTrackerParticipantsText:SetText("Participants (" .. getn(players) .. ")");
		RT_DetailScrollFramePlayers:Show();
	elseif ( frame.type == "items" ) then
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFramePlayer:Hide();
		RT_DetailScrollFrameItems:Show();
		RT_DetailScrollFrameEvents:Hide();
		local numItems, numHidden = 0, 0;
		if raid then
			local keystoremove = {};
		    local loot = {}
			for k, v in pairs(raid.Loot) do
				loot[k] = v;
			end
			table.sort( loot, KA_RaidTracker_CompareItems );
			for key, val in pairs(loot) do
				if (not val.item.tooltip or (getn(val.item.tooltip) == 0)) and db.Options.SaveTooltips == 1 
						and GetItemInfo("item:" .. val.item.id) then
					val.item.tooltip = KARaidTracker:GetItemTooltip(val.item.id);
				end
				if KA_RaidTracker_RarityTable[val.item.c] >= db.SortOptions.itemsfilter then
					numItems = numItems + 1;
				else
					tinsert(keystoremove, key);
					numHidden = numHidden + 1;
				end
			end
			for key, val in pairs(keystoremove) do
				table.remove(loot, val);
			end
			getglobal("RT_DetailScrollFrameItems").raidid = raidid;
			getglobal("RT_DetailScrollFrameItems").loot = loot;
			KA_RaidTracker_DetailScrollFrameItems_Update();
			RT_DetailScrollFrameItems:Show();
		end
		if ( numHidden == 0 ) then
			KA_RaidTrackerItemsText:SetText("Items ("..numItems.."):");
		else
			KA_RaidTrackerItemsText:SetText("Items ("..numItems.."/"..numHidden + numItems..")");
		end
		RT_RarityDropdown:OnUpdate( );
		
	elseif ( frame.type == "player" ) then
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFramePlayer:Show();
		RT_DetailScrollFrameItems:Hide();
		RT_DetailScrollFrameEvents:Hide();
		KA_RaidTrackerPlayerRaidTabLooter:Hide();
		KA_RaidTrackerPlayerRaidTab1:SetWidth(300);
		KA_RaidTrackerPlayerRaidTab1Middle:SetWidth(290);
		local name = frame.player;

		local raids = { };
		for k, v in pairs(db.RaidLog) do
			local isInRaid;
			for key, val in pairs(v.Join) do
				if ( val.player == name ) then
					tinsert(raids, { k, v });
					break;
				end
			end
		end
		
		table.sort(
			raids,
			function(a1, a2)
				if ( db.SortOptions["playerraidway"] == "asc" ) then
					return KarmaUtility:GetTimestamp(a1[2].key) < KarmaUtility:GetTimestamp(a2[2].key);
				else
					return KarmaUtility:GetTimestamp(a1[2].key) > KarmaUtility:GetTimestamp(a2[2].key);
				end
			end
		);
		getglobal("RT_DetailScrollFramePlayer").data = raids;
		getglobal("RT_DetailScrollFramePlayer").name = name;
		getglobal("RT_DetailScrollFramePlayer").maxlines = getn(raids);
		KA_RaidTracker_DetailScrollFramePlayer_Update();
		RT_DetailScrollFramePlayer:Show();
		
		KA_RaidTrackerPlayerText:SetText(name .. "'s Raids (" .. getn(raids) .. "):");
		
	elseif ( frame.type == "itemhistory" ) then
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFramePlayer:Show();
		RT_DetailScrollFrameItems:Hide();
		RT_DetailScrollFrameEvents:Hide();
		KA_RaidTrackerPlayerRaidTabLooter:Show();
		KA_RaidTrackerPlayerRaidTab1:SetWidth(163);
		KA_RaidTrackerPlayerRaidTab1Middle:SetWidth(155);

		local name, totalItems = frame.itemname, 0;

		local loot = { };
		--FIXME: DB Iterator
		for k, v in pairs(db.RaidLog) do
			for key, val in pairs(v.Loot) do
				if ( val.item.name == name ) then
					tinsert(loot, { k, v, val });
					if ( val.item.count ) then
						totalItems = totalItems + val.item.count;
					else
						totalItems = totalItems + 1;
					end
				end
			end
		end
		
		table.sort(
			loot,
			function(a1, a2)
				if ( db.SortOptions["itemhistorymethod"] == "looter" ) then
					if ( db.SortOptions["itemhistoryway"] == "asc" ) then
						return a1[3].player < a2[3].player;
					else
						return a1[3].player > a2[3].player;
					end
				else
					if ( db.SortOptions["itemhistoryway"] == "asc" ) then
						return KarmaUtility:GetTimestamp(a1[2].key) < KarmaUtility:GetTimestamp(a2[2].key);
					else
						return KarmaUtility:GetTimestamp(a1[2].key) > KarmaUtility:GetTimestamp(a2[2].key);
					end
				end
			end
		);

		getglobal("RT_DetailScrollFramePlayer").data = loot;
		getglobal("RT_DetailScrollFramePlayer").name = name;
		getglobal("RT_DetailScrollFramePlayer").maxlines = getn(loot);
		KA_RaidTracker_DetailScrollFramePlayer_Update();
		RT_DetailScrollFramePlayer:Show();
		KA_RaidTrackerPlayerText:SetText(name .. " (" .. getn(loot) .. "/" .. totalItems .. "):");
	elseif ( frame.type == "events" ) then
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFramePlayer:Hide();
		RT_DetailScrollFrameEvents:Show();
		RT_DetailScrollFrameItems:Hide();
		KA_RaidTrackerPlayerBossesTabBoss:Show();
		KA_RaidTrackerPlayerBossesTab1:SetWidth(163);
		KA_RaidTrackerPlayerBossesTab1Middle:SetWidth(155);

		local events = {};
		if ( raid and raid.BossKills ) then
			for key, val in pairs(raid.BossKills) do
				tinsert(events, val);
			end
		end
		getglobal("RT_DetailScrollFrameEvents").raidid = raidid;
		getglobal("RT_DetailScrollFrameEvents").events = events;
		
		KA_RaidTrackerEventsText:SetText("Events");
		KA_RaidTracker_DetailScrollFrameBoss_Update();
		RT_DetailScrollFrameEvents:Show();
		
	elseif ( frame.type == "playeritems" ) then
		RT_DetailScrollFramePlayers:Hide();
		RT_DetailScrollFramePlayer:Hide();
		RT_DetailScrollFrameItems:Show();
		RT_DetailScrollFrameEvents:Hide();
		local name = frame.player;

		local loot = { };
		for k, v in pairs(db.RaidLog) do
			for key, val in pairs(v.Loot) do
				if ( val.player == name ) then
					tinsert(
						loot,
						{
							["note"] = val.note,
							["player"] = val.player,
							["time"] = val.time,
							item = val.item,
							["ids"] = { k, key }
						}
					);
				end
			end
		end

		local numItems, numHidden = 0, 0;
		local keystoremove = {};
		table.sort( loot, KA_RaidTracker_CompareItems );
		for key, val in pairs(loot) do
			if ( KA_RaidTracker_RarityTable[val.item.c] >= db.SortOptions["playeritemsfilter"] ) then
				numItems = numItems + 1;
			else
				tinsert(keystoremove, key);
				numHidden = numHidden + 1;
			end
		end
		for key, val in pairs(keystoremove) do
			table.remove(loot, val);
		end
		getglobal("RT_DetailScrollFrameItems").raidid = raidid;
		getglobal("RT_DetailScrollFrameItems").loot = loot;
		KA_RaidTracker_DetailScrollFrameItems_Update();
		if ( numHidden == 0 ) then
			KA_RaidTrackerItemsText:SetText(name .. "'s Loot (" .. numItems .. "):");
		else
			KA_RaidTrackerItemsText:SetText(name .. "'s Loot (" .. numItems .. "/" .. numHidden + numItems .. "):");
		end

		RT_RarityDropdown:OnUpdate( );
	end
end

function KA_RaidTracker_DetailScrollFramePlayers_Update()
	local raidid = getglobal("RT_DetailScrollFramePlayers").raidid;
	local players = getglobal("RT_DetailScrollFramePlayers").players;
	local maxlines = getn(players);
	local line;
	local lineplusoffset;
	FauxScrollFrame_Update(RT_DetailScrollFramePlayers, maxlines, 11, 18);
	for line=1, 11 do
		lineplusoffset = line+FauxScrollFrame_GetOffset(RT_DetailScrollFramePlayers);
		if (lineplusoffset <= maxlines) then
			val = players[lineplusoffset];
			getglobal("KA_RaidTrackerPlayerLine" .. line).raidid = raidid;
			getglobal("KA_RaidTrackerPlayerLine" .. line).raidtitle = KA_RaidTracker_GetRaidTitle(raidid, 1);
			getglobal("KA_RaidTrackerPlayerLine" .. line).playername = val.name;
			local name = getglobal("KA_RaidTrackerPlayerLine" .. line .. "Name");
			local number = getglobal("KA_RaidTrackerPlayerLine" .. line .. "Number");
			local join = getglobal("KA_RaidTrackerPlayerLine" .. line .. "Join");
			local leave = getglobal("KA_RaidTrackerPlayerLine" .. line .. "Leave");
			if ( name ) then
				name:SetText(val.name);
				local iNumber = lineplusoffset;
				if ( iNumber < 10 ) then
					iNumber = "  " .. iNumber;
				end
				number:SetText(iNumber);
				if(KARaidTrackerDB.Options["TimeFormat"] == 1) then
					join:SetText(date("%H:%M", val["join"]));
				else
					join:SetText(date("%I:%M%p", val["join"]));
				end
				if ( val["leave"] == 99999999999 ) then
					leave:SetText("");
				else
					if(KARaidTrackerDB.Options["TimeFormat"] == 1) then
						leave:SetText(date("%H:%M", val["leave"]));
					else
						leave:SetText(date("%I:%M%p", val["leave"]));
					end
				end
				if ( KARaidTrackerDB.RaidLog[raidid].PlayerInfos[val.name] and KARaidTrackerDB.RaidLog[raidid].PlayerInfos[val.name].note ) then
					getglobal("KA_RaidTrackerPlayerLine" .. line .. "NoteButtonNormalTexture"):SetVertexColor(1, 1, 1);
				else
					getglobal("KA_RaidTrackerPlayerLine" .. line .. "NoteButtonNormalTexture"):SetVertexColor(0.5, 0.5, 0.5);
				end
				getglobal("KA_RaidTrackerPlayerLine" .. line .. "NoteButton"):Show();
				getglobal("KA_RaidTrackerPlayerLine" .. line .. "DeleteButton"):Show();
				getglobal("KA_RaidTrackerPlayerLine" .. line):Show();
			end
		else
			getglobal("KA_RaidTrackerPlayerLine" .. line):Hide();
		end
	end
	RT_DetailScrollFramePlayers:Show();
end

function KA_RaidTracker_DetailScrollFrameItems_Update()
	local frame = RaidTrackerFrame; 
	local db = KARaidTrackerDB;	

	local raidid = getglobal("RT_DetailScrollFrameItems").raidid;
	local loot = getglobal("RT_DetailScrollFrameItems").loot;
	local raid = db.RaidLog[raidid]	

	local numEntries = getn(loot);
	FauxScrollFrame_Update(RT_DetailScrollFrameItems, numEntries, 5, 41);
	for i=1, 5 do
		local line = getglobal("KA_RaidTrackerItem" .. i);
		local itemid = i + FauxScrollFrame_GetOffset(RT_DetailScrollFrameItems);
		
		if itemid > numEntries then
			line:Hide();
		else
			local item = loot[itemid]
			
			if frame.type == "items" then
				line.raidid = raidid;
				line.itemid = item.id;
				line.itemname = item.item.name;
				if item.item.count and item.item.count > 1 then
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):Show();
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):SetText(item.item.count);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):Hide();
				end
				if item.item.icon then
					getglobal("KA_RaidTrackerItem" .. i .. "IconTexture"):SetTexture("Interface\\Icons\\" .. item.item.icon);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "IconTexture"):SetTexture("Interface\\Icons\\INV_Misc_Gear_08");
				end
				local color = item.item.c;
				if color == "ff1eff00" then
					color = "ff005D00";
				end
				getglobal("KA_RaidTrackerItem" .. i .. "Description"):SetText("|c" .. color .. item.item.name);
				getglobal("KA_RaidTrackerItem" .. i .. "Looted"):SetText("Looted by: " .. KARaidTracker:ToStr(item.player));
	
				if item.note then
					getglobal("KA_RaidTrackerItem" .. i .. "NoteNormalTexture"):SetVertexColor(1, 1, 1);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "NoteNormalTexture"):SetVertexColor(0.5, 0.5, 0.5);
				end
			elseif frame.type == "playeritems" then
				line.raidid = item.ids[1];
				line.itemid = item.ids[2];
				line.itemname = item.item.name;

				if ( item.item.count and item.item.count > 1 ) then
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):Show();
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):SetText(item.item.count);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "Count"):Hide();
				end
				if ( item.item.icon ) then
					getglobal("KA_RaidTrackerItem" .. i .. "IconTexture"):SetTexture("Interface\\Icons\\" .. item.item.icon);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "IconTexture"):SetTexture("Interface\\Icons\\INV_Misc_Gear_08");
				end
				local color = item.item.c;
				if ( color == "ff1eff00" ) then
					color = "ff005D00";
				end
				getglobal("KA_RaidTrackerItem" .. i .. "Description"):SetText("|c" .. color .. item.item.name);
				getglobal("KA_RaidTrackerItem" .. i .. "Looted"):SetText("Looted " .. KA_RaidTracker_GetRaidTitle(item.ids[1], 1));

				if ( item.note ) then
					getglobal("KA_RaidTrackerItem" .. i .. "NoteNormalTexture"):SetVertexColor(1, 1, 1);
				else
					getglobal("KA_RaidTrackerItem" .. i .. "NoteNormalTexture"):SetVertexColor(0.5, 0.5, 0.5);
				end	
			end
			line:Show();
		end
	end
	RT_DetailScrollFrameItems:Show();
end

function KA_RaidTracker_DetailScrollFramePlayer_Update()
	local data = getglobal("RT_DetailScrollFramePlayer").data;
	local name = getglobal("RT_DetailScrollFramePlayer").name;
	local maxlines = getglobal("RT_DetailScrollFramePlayer").maxlines;
	local line;
	local lineplusoffset;
	FauxScrollFrame_Update(RT_DetailScrollFramePlayer, maxlines, 11, 18);
	for line=1, 11 do
		lineplusoffset = line+FauxScrollFrame_GetOffset(RT_DetailScrollFramePlayer);
		if (lineplusoffset <= maxlines) then
			val = data[lineplusoffset];
			if ( RaidTrackerFrame.type == "player" ) then
				getglobal("KA_RaidTrackerPlayerRaid" .. line).raidid = val[1];
				getglobal("KA_RaidTrackerPlayerRaid" .. line).playername = name;
				getglobal("KA_RaidTrackerPlayerRaid" .. line).raidtitle = KA_RaidTracker_GetRaidTitle(val[1], 1);
	
				local iNumber = getn(KARaidTrackerDB.RaidLog)-val[1]+1;
				if ( iNumber < 10 ) then
					iNumber = "  " .. iNumber;
				end
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "MouseOverLeft"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "MouseOverRight"):Hide();
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitAreaLeft"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitAreaRight"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitArea"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Note"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "NoteButton"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "DeleteButton"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "DeleteText"):Show();
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Number"):SetText(iNumber);
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Name"):SetWidth(200);
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Name"):SetText(KA_RaidTracker_GetRaidTitle(val[1], 1));

				if ( val[2].PlayerInfos[name] and val[2].PlayerInfos[name].note ) then
					getglobal("KA_RaidTrackerPlayerRaid" .. line .. "NoteButtonNormalTexture"):SetVertexColor(1, 1, 1);
				else
					getglobal("KA_RaidTrackerPlayerRaid" .. line .. "NoteButtonNormalTexture"):SetVertexColor(0.5, 0.5, 0.5);
				end
			elseif ( RaidTrackerFrame.type == "itemhistory" ) then
				getglobal("KA_RaidTrackerPlayerRaid" .. line).raidid = val[1];
	
				local iNumber = getn(KARaidTrackerDB.RaidLog)-val[1]+1;
				if ( iNumber < 10 ) then
					iNumber = "  " .. iNumber;
				end
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "MouseOver"):Hide();
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitAreaLeft"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitAreaRight"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "HitArea"):Hide();
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "NoteButton"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Note"):Show();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "DeleteButton"):Hide();
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "DeleteText"):Hide();
	
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Number"):SetText(iNumber);
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Name"):SetWidth(130);
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Name"):SetText(KA_RaidTracker_GetRaidTitle(val[1], 1));
				getglobal("KA_RaidTrackerPlayerRaid" .. line .. "Note"):SetText(val[3].player);
			end
			getglobal("KA_RaidTrackerPlayerRaid" .. line):Show();
		else
			getglobal("KA_RaidTrackerPlayerRaid" .. line):Hide();
		end
	end
	RT_DetailScrollFramePlayer:Show();
end

function KA_RaidTracker_DetailScrollFrameBoss_Update()
	local events = getglobal("RT_DetailScrollFrameEvents").events;
	local maxlines = getn(events);
	local line;
	local lineplusoffset;
	FauxScrollFrame_Update(RT_DetailScrollFrameEvents, maxlines, 11, 18);
	for line=1, 11 do
		lineplusoffset = line+FauxScrollFrame_GetOffset(RT_DetailScrollFrameEvents);
		if (lineplusoffset <= maxlines) then
			val = events[lineplusoffset];
			getglobal("KA_RaidTrackerBosses" .. line .. "MouseOver"):Hide();
			getglobal("KA_RaidTrackerBosses" .. line .. "HitArea"):Show();
			getglobal("KA_RaidTrackerBosses" .. line .. "Boss"):SetText(val.boss);
			getglobal("KA_RaidTrackerBosses" .. line .. "Time"):SetText(val.time);
			getglobal("KA_RaidTrackerBosses" .. line):Show();
		else
			getglobal("KA_RaidTrackerBosses" .. line):Hide();
		end
	end
	RT_DetailScrollFrameEvents:Show();
end


function KA_RaidTrackerItem_SetHyperlink( frame )
	local db = KARaidTrackerDB
	if not (frame and frame.raidid and frame.itemid) then return end
	local item = db.RaidLog[frame.raidid].Loot[frame.itemid]
	
	GameTooltip:SetOwner(frame, "ANCHOR_RIGHT");
	if GetItemInfo("item:" .. item.item.id) then
		GameTooltip:SetHyperlink("item:" .. item.item.id);
	else
		rl, gl, bl = KARaidTracker:ColorToRGB(strsub(item.item.c, 3));
		GameTooltip:AddLine(item.item.name, rl, gl, bl);
		GameTooltip:AddLine("This item is not in your cache, you can try to get the information by rightclicking the item (This may result in a disconnect!)", 1, 1, 1);
	end
	
	GameTooltip:AddLine("Time: "..item.time, 1, 1, 0);
	if item.zone then	GameTooltip:AddLine("Zone: "..item.zone, 1, 1, 0)	end
	if item.boss then	GameTooltip:AddLine("Boss: "..item.boss, 1, 1, 0)	end
	if item.costs then	GameTooltip:AddLine("Costs: "..item.costs, 1, 1, 0) end
	if db.Options.DebugMode == 1 then
		if item.item.class then		GameTooltip:AddLine("Class: "..item.item.class, 1, 1, 0) end
		if item.item.subclass then	GameTooltip:AddLine("SubClass: "..item.item.subclass, 1, 1, 0) end
	end
	
	GameTooltip:Show();
end

function KA_RaidTrackerItem_GetChatHyperlink( frame )
	local db = KARaidTrackerDB
	if not (frame and frame.raidid and frame.itemid) then return end
	local item = db.RaidLog[frame.raidid].Loot[frame.itemid];
	
	if item and IsShiftKeyDown() and type(WIM_API_InsertText)=="function" then
		WIM_API_InsertText("|c"..item.item.c.."|Hitem:"..item.item.id.."|h["..item.item.name.."]|h|r");
	end	
	if item and IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		ChatFrameEditBox:Insert("|c"..item.item.c.."|Hitem:"..item.item.id.."|h["..item.item.name.."]|h|r");
	end
end


function KA_RaidTracker_AddWipe()
	local db = KARaidTrackerDB
	KarmaUtility:Debug("WIPED");
	KA_RaidTracker_LastWipe = GetTime()+db.Options.WipeCooldown; -- wait for 120 seconds
	if db.RaidLog[db.Options.CurrentRaid].wipes == nil then
		db.RaidLog[db.Options.CurrentRaid].wipes = {};
	end;
	tinsert(db.RaidLog[db.Options.CurrentRaid].wipes,time());
	KA_RaidTracker_Print("Wipe has been recorded!", 1, 1, 0);
end;

function KA_RaidTracker_NextBoss(name)
	KarmaUtility:Debug("NEXTBOSS",name);
	KARaidTrackerDB.RaidLog[KARaidTrackerDB.Options["CurrentRaid"]].nextboss = name;
end;

function KA_RaidTracker_DoRaidIdCheck()
	local db = KARaidTrackerDB
	if not db.Options.CurrentRaid then return end
	
	local savedInstances = GetNumSavedInstances();
	local instanceName, instanceID, instanceReset;
	if ( savedInstances > 0 ) then
		for i=1, MAX_RAID_INFOS do
			if ( i <=  savedInstances) then
				instanceName, instanceID, instanceReset = GetSavedInstanceInfo(i);
				if (db.RaidLog[db.Options.CurrentRaid].zone == instanceName) then
					db.RaidLog[db.Options.CurrentRaid].instanceid = instanceID;
					--db.RaidLog[db.Options.CurrentRaid]["instancereset"] = instanceReset;
				end;
			end
			
		end
	end
end;


function KA_RaidTracker_SelectItem(name)
	KA_RaidTracker_GetPage();
	RaidTrackerFrame.type = "itemhistory";
	RaidTrackerFrame.itemname = name;
	RaidTrackerFrame.selected = nil;
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end

function KA_RaidTracker_GetPage()
	if ( RaidTrackerFrame.type or RaidTrackerFrame.itemname or RaidTrackerFrame.selected or RaidTrackerFrame.player ) then

		tinsert(RT_Buttons.LastPage,
			{
				["type"] = RaidTrackerFrame.type,
				["itemname"] = RaidTrackerFrame.itemname,
				["selected"] = RaidTrackerFrame.selected,
				["player"] = RaidTrackerFrame.player
			}
		);
	end

	if ( getn(RT_Buttons.LastPage) > 0 ) then
		RaidTrackerFrameBackButton:Enable();
	else
		RaidTrackerFrameBackButton:Disable();
	end
end

function KA_RaidTracker_GoBack()
	local t = table.remove(RT_Buttons.LastPage);

	if ( t ) then
		RaidTrackerFrame.type = t["type"];
		RaidTrackerFrame.itemname = t["itemname"];
		RaidTrackerFrame.selected = t["selected"];
		RaidTrackerFrame.player = t.player;
		KA_RaidTracker_Update();
		KA_RaidTracker_UpdateView();
	end
	if ( getn(RT_Buttons.LastPage) > 0 ) then
		RaidTrackerFrameBackButton:Enable();
	else
		RaidTrackerFrameBackButton:Disable();
	end
end

function KA_RaidTrackerSnapshotRaid()
    local db = KARaidTrackerDB
    local sDate = KARaidTracker:GetDisplayTime();
    local newraid = {};
    if(db.Options.CurrentRaid) then
        KarmaUtility:Print("Snapshotting current raid");
        tinsert(db.RaidLog, 2, { 
    		["Loot"] = { },
    		["Join"] = { },
    		["Leave"] = { },
    		["PlayerInfos"] = { },
    		["BossKills"] = { },
    		["key"] = sDate,
    		["End"] = sDate,
	        });
	    if db.RaidLog[db.Options.CurrentRaid].zone then
	        db.RaidLog[2].zone = db.RaidLog[db.Options.CurrentRaid].zone;
	    end
	    
	    for k, v in pairs(db.Online) do
            db.RaidLog[2].PlayerInfos[k] = {};
            tinsert(db.RaidLog[2].Join, {
                ["player"] = k,
                ["time"] = sDate 
                });
            tinsert(db.RaidLog[2].Leave, {
                ["player"] = k,
                ["time"] = sDate 
                });
	    end
	    KA_RaidTracker_UpdateView();
    	KA_RaidTracker_Update();
    end
end
	
function KA_RaidTrackerAddGuild()
    local db = KARaidTrackerDB

	if(db.Options.CurrentRaid) then
		SetGuildRosterShowOffline(false);
     for i = 1, GetNumGuildMembers() do
        local name, rank, rankIndex, level, class, zone, group, note, officernote, online = GetGuildRosterInfo(i);
       -- KarmaUtility:Debug("GUILD", name, online);
        if( online ~= db.Online[name] ) then
          if( online ) then
            tinsert(db.RaidLog[db.Options.CurrentRaid].Join,
            {
               ["player"] = name,
               ["time"] = KARaidTracker:GetDisplayTime()
            }
            );
            db.Online[name] = online;
            KarmaUtility:Debug("GUILD-ONLINE", name);
          elseif ( not online and db.Online[name]) then
						tinsert(db.RaidLog[db.Options.CurrentRaid].Leave,
							{
								["player"] = name,
								["time"] = KARaidTracker:GetDisplayTime()
							}
						);
						db.Online[name] = online;
						KarmaUtility:Debug("GUILD-OFFLINE", name);
         	end
       	end
     end
  end
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();
end


function KA_RaidTracker_EditNote_OnShow( frame )
    local db = KARaidTrackerDB
	local hTitle, hText, text;
	local item, raid;
	if not frame then return end
		
	raid = db.RaidLog[frame.raidid]
	if frame.itemid then
		item = raid.Loot[frame.itemid]
	end

	if frame.type == "raidnote" then
	    hText = "Editing note for \"|c" .. "0000ff00" .. raid.key .. "|r\"";
		text = raid.note;		
	elseif frame.type == "itemnote" then
	    hText = "Editing note for \"|c" .. item.item.c .. "" .. item.item.name .. "|r\"";
		text = item.note;
	elseif frame.type == "itemcount" then
	    hTitle = "Edit Count";
	    hText = "Editing count for \"|c" .. item.item.c .. "" .. item.item.name .. "|r\"";
		text = item.item.count;
	elseif frame.type == "itemcost" then
	    hTitle = "Edit Cost";
        hText = "Editing cost for \"|c" .. item.item.c .. "" .. item.item.name .. "|r\"";
        text = tostring(item.costs or "");
	elseif frame.type == "playernote" then
	    hText = "Editing note for player \"" .. frame.playerid .. "\"";
		text = raid.PlayerInfos[frame.playerid] and raid.PlayerInfos[frame.playerid].note;
	elseif frame.type == "looter" then
	    hTitle = "Edit Looter";
	    hText = "Editing looter for \"|c" .. item.item.c .. "" .. item.item.name .. "|r\"";
	    text = item.player;
	elseif frame.type == "time" then
		hTitle = "Edit Time";
		if frame.what == "raidend" then
			hText = "Editing End Time";
			text = raid.End;
		elseif frame.what == "raidstart" then
			hText = "Editing Start Time";
			text = raid.key;
		elseif frame.what == "item" then
			hText = "Editing Item Time";
			text = item.time;
		end
	elseif frame.type == "zone" then
	    hTitle = "Edit Zone";
	    hText = "Editing zone for \"|c" .. "0000ff00" .. "" .. raid.key .. "|r\"";
	    text = raid.zone;
	elseif frame.type == "export" then
	    hTitle = "Export String";
	    hText = "Showing "..RaidTracker_Options.desc.export[db.Options.ExportFormat].." string for upload";
	    text = RaidTracker_Export:GenerateRaid(raid);	
	end
	

	if hTitle then getglobal(frame:GetName().."HeaderTitle"):SetText(hTitle); end
	if hText then getglobal(frame:GetName().."HeaderText"):SetText(hText); end
	getglobal(frame:GetName().."TextBox"):SetText(text or "");
	getglobal(frame:GetName().."TextBox"):HighlightText();

	KARaidTracker:Debug("KA_RaidTracker_EditNote_OnShow",frame:GetName(),frame.raidid,hTitle,hText);
end


function KA_RaidTracker_SaveNote( frame, option )
    local db = KARaidTrackerDB
	local text, item, raid;
	if not frame then return end
		
	raid = db.RaidLog[frame.raidid]
	if frame.itemid then
		item = raid.Loot[frame.itemid]
	end	
	text = getglobal(frame:GetName().."TextBox"):GetText()
	if strlen(text) == 0 then
		text = nil;
	end
	
	if frame.type == "itemnote" then
		item.note = text;
	elseif frame.type == "itemcount" then
		if not text or not tonumber(text) then
			KARaidTracker:Print("Edit Count: Invalid value.");
		else
			item.item.count = tonumber(text);
		end
	elseif frame.type == "itemcost" then
		if text and (not tonumber(text) or not string.find(text, "^(%d+%.?%d*)$")) then
    		KARaidTracker:Print("Edit Costs: Invalid value");
		else
  			item.costs = text
  			if type(dkpp_ctra_sub) == "function" then
  				dkpp_ctra_sub(frame.raidid, frame.itemid);
			end;
		end
		if option == "bank" or option=="disenchant" then item.player = option end
	elseif frame.type == "raidnote" then
		raid.note = text;
	elseif frame.type == "playernote" then
		if not raid.PlayerInfos[frame.playerid] then
			raid.PlayerInfos[frame.playerid] = {};
		end
		raid.PlayerInfos[frame.playerid].note = text;
	elseif frame.type == "looter" then
	    if text then
			item.player = text;
	    end
	elseif frame.type == "time" then
	    if text and not string.find(text, "^(%d+)/(%d+)/(%d+) (%d+):(%d+):(%d+)$") then
		    KarmaUtility:Print("RaidTracker Edit Time: Invalid Time format");
		  else
		  	if		frame.what == "raidend" then	raid.End = text;
		    elseif	frame.what == "raidstart" then	raid.key = text;
			elseif	frame.what == "item" then		item.time = text;
		    end
	    end
	elseif frame.type == "zone" then
	    KA_RaidTracker_SaveZone(frame.raidid, text);
	end

	KARaidTracker:Debug("KA_RaidTracker_SaveNote",frame:GetName(),frame.raidid,option);
	
	frame.type = nil;
	frame.raidid = nil;
	frame.playerid = nil;
	frame.what = nil;
	frame.itemid = nil;
	
	KA_RaidTracker_Update();
	KA_RaidTracker_UpdateView();	
end


function KA_RaidTracker_SaveZone(raidid, text)
    local db = KARaidTrackerDB

	local zone, zonetrigger, zonefound;
	if (text == nil or strlen(text) == 0 ) then
		text = nil;
		zone = nil;
		zonetrigger = nil;
	elseif( string.find(text, "^(.+)%-(.+)$") ) then
		_, _, zone, zonetrigger = string.find(text, "^(.+)%-(.+)$");
	else
		zone = text;
		zonetrigger = text;
	end
	if(zone and zonetrigger) then
		if(not db.RaidLog[raidid].zone or db.RaidLog[raidid].zone ~= zone) then
			for k, v in pairs(RaidTracker_ZoneTriggers) do
				if(zonetrigger == k) then
					zonefound = 1;
					break;
				end
			end
			if(not zonefound) then
				for k, v in pairs(db.ZoneTriggers) do
					if(zonetrigger == k) then
						zonefound = 1;
						break;
					end
				end
			end
			if(not zonefound) then
				KarmaUtility:Print("KA_RaidTracker Custom Zones: Added \""..zone.."\" (Trigger: \""..zonetrigger.."\")");
				db.ZoneTriggers[zonetrigger] = zone;
			end
		end
	elseif(not zone and not zonetrigger and db.RaidLog[raidid].zone) then
		for k, v in pairs(db.ZoneTriggers) do
			if(v == db.RaidLog[raidid].zone) then
				db.ZoneTriggers[k] = nil;
				KarmaUtility:Print("KA_RaidTracker Custom Zones: Removed \""..v.."\" (Trigger: \""..k.."\")");
			end
		end
	end
	
	db.RaidLog[raidid].zone = zone;
end

function KA_RaidTracker_JoinLeaveSave()
    local db = KARaidTrackerDB

    local player_name = RT_JoinLeaveFrameNameEB:GetText();
	local player_note = RT_JoinLeaveFrameNoteEB:GetText();
	local player_time = RT_JoinLeaveFrameTimeEB:GetText();
	
	if(player_name == nil or strlen(player_name) == 0) then
	    KarmaUtility:Print("KA_RaidTracker Join/Leave: No player");
	    return nil;
	end
	if(not string.find(player_time, "^(%d+)/(%d+)/(%d+) (%d+):(%d+):(%d+)$")) then
	    KarmaUtility:Print("KA_RaidTracker Join/Leave: Invalid Time format");
	    return nil;
    end
	
    if((strlen(player_name) > 0)) then
        local sDate = KARaidTracker:GetDisplayTime();
        KarmaUtility:Debug("KA_RaidTracker_JoinLeave", player_name, player_note);
        if (RT_JoinLeaveFrame.type == "Join") then 
            tinsert(db.RaidLog[RT_JoinLeaveFrame.raidid].Join,
            {
               ["player"] = player_name,
               time = player_time
            }
            );
            db.Online[player_name] = 1;
            KarmaUtility:Print(player_name.." manually joined at "..player_time);
        elseif (RT_JoinLeaveFrame.type == "Leave") then
            tinsert(db.RaidLog[RT_JoinLeaveFrame.raidid].Leave,
            {
               ["player"] = player_name,
               ["time"] = player_time
            }
            );
            db.Online[player_name] = nil;
            KarmaUtility:Print(player_name.." manually left at "..player_time);
        end
        if(strlen(player_note) > 0) then
            if( not db.RaidLog[RT_JoinLeaveFrame.raidid].PlayerInfos[player_name]) then
	            db.RaidLog[RT_JoinLeaveFrame.raidid].PlayerInfos[player_name] = {};
            end
            db.RaidLog[RT_JoinLeaveFrame.raidid].PlayerInfos[player_name].note = player_note;
        end
        KA_RaidTracker_Update();
        KA_RaidTracker_UpdateView();
    end
end


-- Next Boss selection handling

function KA_RaidTracker_Boss_InitWindow()
	UIDropDownMenu_Initialize(RT_NextBossFrameBossDropdown, KA_RaidTracker_Boss_Init);
end;

function KA_RaidTracker_Boss_Init()
    local db = KARaidTrackerDB

	local i = 0;
	local ii = 0;
	if (db.Options.CurrentRaid == nil) then
		return;
	end;

	i = KA_RaidTracker_Boss_Add_Button(RaidTracker_BossUnitTriggers["DEFAULTBOSS"],i);
	ii = ii + 1;
	if (db.RaidLog[db.Options.CurrentRaid].zone == nil) then
		for k,v in pairs(RaidTracker_Bosses) do
				KarmaUtility:Debug(k,v);
			if (v == 1) then
				ii = ii + 1;
				for k2,v2 in pairs(RaidTracker_BossUnitTriggers) do
					if (v2 == v) then
						i = KA_RaidTracker_Boss_Add_Button(k2,i);
					end;
				end;
			end;
		end
	else
		for k,v in pairs(RaidTracker_Bosses[db.RaidLog[db.Options.CurrentRaid].zone]) do
			local addit = true;
			for k2,v2 in pairs(db.RaidLog[db.Options.CurrentRaid].BossKills) do
				if (v2.boss == v) then
					addit = false;
				end;
			end;
			if (addit == true) then
				if (type(v) == "table") then
					for k2,v2 in pairs(v) do
						ii = ii + 1;
						for k3,v3 in pairs(RaidTracker_BossUnitTriggers) do
							if (v3 == v2) then
								i = KA_RaidTracker_Boss_Add_Button(k..' - '..k3,i);
							end;
						end;
					end;
				else
					ii = ii + 1;
					for k2,v2 in pairs(RaidTracker_BossUnitTriggers) do
						if (v2 == v) then
							i = KA_RaidTracker_Boss_Add_Button(k2,i);
						end;
					end;
				end;
			end;
		end
	end;
	if (ii == 2) then
		UIDropDownMenu_SetSelectedID(RT_NextBossFrameBossDropdown, ii);
	end;
end;

function KA_RaidTracker_Boss_Add_Button(k,i)
	local info = {
		text = k;
		func = KA_RaidTracker_Boss_Update;
	};
	UIDropDownMenu_AddButton(info);
	if (i == cur_class_id) then
		UIDropDownMenu_SetSelectedID(RT_NextBossFrameBossDropdown, i+1);
		UIDropDownMenu_SetText(info.text, RT_NextBossFrameBossDropdown);
	end
	return i
end;

function KA_RaidTracker_Boss_Update( )
	i = this:GetID();
	UIDropDownMenu_SetSelectedID(RT_NextBossFrameBossDropdown, i);
end;


-- Time offset handling

function KA_RaidTracker_GameTimeFrame_Update()
	-- call orginal blizz function
	KA_RaidTracker_GameTimeFrame_Update_Original();
	local db = KARaidTrackerDB;
	
	local hour, minute = GetGameTime();
	local time = ((hour * 60) + minute) * 60;
	
	if( not db.Options["TimeOffsetStatus"] ) then
		db.Options["TimeOffsetStatus"] = time;
	elseif ( db.Options["TimeOffsetStatus"] ~= time ) then
		local ltimea = date("*t");
		local ltime = (((ltimea["hour"] * 60) + ltimea["min"]) * 60 + ltimea["sec"]) + (db.Options["Timezone"] * 3600);

		if( time > ltime ) then
			local timediff = time - ltime;
			if(timediff >= 43200) then
				KA_RaidTracker_TimeOffset = timediff - 86400;
			else
				KA_RaidTracker_TimeOffset = timediff;
			end
		elseif ( time < ltime ) then
			local timediff = ltime - time;
			if(timediff >= 43200) then
				KA_RaidTracker_TimeOffset = 86400 - timediff;
			else
				KA_RaidTracker_TimeOffset = timediff * -1;
			end
		else
			KA_RaidTracker_TimeOffset = 0;
		end

		KarmaUtility:Debug("KA_RaidTracker_TimeOffset", KA_RaidTracker_TimeOffset);
		GameTimeFrame_Update = KA_RaidTracker_GameTimeFrame_Update_Original;
		db.Options["TimeOffsetStatus"] = nil;
	end
end

function KA_RaidTracker_GetGameTimeOffset()
	if KARaidTrackerDB.Options["TimeOffsetStatus"] then
		return;
	end
	-- hook wow GameTimeFrame_Update function
	if (not KA_RaidTracker_GameTimeFrame_Update_Original) then
		KA_RaidTracker_GameTimeFrame_Update_Original = GameTimeFrame_Update;
	end
	GameTimeFrame_Update = KA_RaidTracker_GameTimeFrame_Update;
end

function RT_UpdateFrame_OnUpdate( frame, elapsed )
	if ( frame.time ) then
		frame.time = frame.time + elapsed;
		if ( frame.time > 2 ) then
			frame.time = nil;
			for k, v in pairs(KA_RaidTracker_Events) do
				KARaidTracker:Debug("Running", unpack(v));
				KARaidTracker:OnEvent(unpack(v));
			end
		end
	end
end

